/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.persistence.api;

import org.objectweb.perseus.cache.api.CacheEntry;

/**
 *
 * @author Luciano Garcia-Banuelos (Luciano.Garcia@imag.fr)
 */
public interface PersistenceManager {
	/**
     * Given an object, this method creates the corresponding copy in
     * permanent support.  At the same time, it  establishes a binding
     * between the object in main-memory and the permanent copy.
     * Roughly, the internal tasks include:
     * - Reserving space on  permanent  support,
     * - Establishing a binding between the object  in main memory and
     * the permanent copy,
     * - Marking the object as dirty, (internal  or by the concurrency
     * control module?)
     * - Adding the object to the cache
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
     * @param  obj The  object  whose state  is  to  be copied  to
     * @return the state associated to the working set
     * @exception PersistenceException Whenever the object is invalid
	 * (e.g. NULL) or it has been previously bound to another oid.
     */
	State export(WorkingSet context, Object obj) throws PersistenceException;

    /**
	 * This method permits to given some information to build the identifier of
	 * the new persistent object.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param  obj The  object  whose state  is  to  be copied  to
	 * @param hints can be used to build the identifier of the object.
     * @return the state associated to the working set
     * @exception PersistenceException Whenever the object is invalid
	 * (e.g. NULL) or it has been previously bound to another oid.
	 */
	State export(WorkingSet context, Object obj, Object hints) throws PersistenceException;

	/**
     * Deletes the binding between  an in main-memory object, and its
     * corresponding copy  in permanent support. At the  same time, it
     * frees (may be) the  space in the permanent support allocated to
     * the permanent copy.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param  oid The  identifier of the  corresponding object  to be
	 * destroyed.
     * @return the state associated to the working set
     * @exception PersistenceException
     */
    State unexport(WorkingSet context, Object oid) throws PersistenceException;

	/**
     * Deletes the binding between  an in main-memory object, and its
     * corresponding copy  in permanent support. At the  same time, it
     * frees (may be) the  space in the permanent support allocated to
     * the permanent copy.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param  ce The  identifier of the  corresponding object  to be
	 * destroyed.
     * @return the state associated to the working set
     * @exception PersistenceException
     */
	State unexport(WorkingSet context, CacheEntry ce) throws PersistenceException;

	/**
	 * This method retrieves a CacheEntry without checking if the persistent 
	 * object, exists. If the entry is already in the working or in the cache, 
	 * the found entry is returned. Otherwise a new CacheEntry is returned.
	 * This method does not permit an access to persistent data, because no
	 * intention has been declared. Then a call to readIntention or 
	 * writeIntention is required.
	 * 
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param  oid The  identifier of the  corresponding object to find
	 */	
	CacheEntry getObjectById(WorkingSet context, Object oid) throws PersistenceException;
	
    /**
	 * This method records an access intention to a data object in read mode.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param oid is the object identifier.
     * @return the state associated to the working set
	 * @exception PersistenceException
	 */
	State readIntention(WorkingSet context, Object oid, Object thinLock) throws PersistenceException;

    /**
	 * This method records an access intention to a data object in read mode.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param ce is the cache entry.
     * @return the state associated to the working set
	 * @exception PersistenceException
	 */
	State readIntention(WorkingSet context, CacheEntry ce, Object thinLock) throws PersistenceException;

	/**
	 * This method records an access intention to a data object in write mode.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param oid is the object identifier.
     * @return the state associated to the working set
	 * @exception PersistenceException
	 */
    State writeIntention(WorkingSet context, Object oid, Object thinLock) throws PersistenceException;

	/**
	 * This method records an access intention to a data object in write mode.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param ce is the cache entry.
     * @return the state associated to the working set
	 * @exception PersistenceException
	 */
    State writeIntention(WorkingSet context, CacheEntry ce, Object thinLock) throws PersistenceException;

	/**
	 * This method can be used to mark the end of an access to a data object.
	 * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param ce is the  cache entry on which the access is finished.
	 * @exception PersistenceException
	 */
    void accessCompletion(WorkingSet context, State ce) throws PersistenceException;

    /**
     * This method permits to flush entries associated to a context.
     * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
     * @param statefilter is a filter of the cache entry which must be flushed.
     * A null StateFilter means all entries associated to the context has to be
     * flushed. 
     * @throws PersistenceException if the context is not valid.
     */
    void flush(WorkingSet context, StateFilter statefilter) throws PersistenceException;

    /**
     * This method permits to flush entries associated to a context.
     * @param context is an identifier of the execution context. It can be a
	 * transaction handle.
     * @param state is the state taht must be flushed
     * @throws PersistenceException if the context is not valid.
     */
    void flush(WorkingSet context, State state) throws PersistenceException;
    
	/**
	 * This methods permits to force the cache eviction of a persistent object
	 * designed by its identifier.
	 * @param context is the working set asking the eviction.
	 * @param oid is the identifier of the persistent object to evict from the
	 * cache
	 * @param force indicates if the entry MUST be evicted (true) or TRIES to be
	 * evicted (false).
	 * @return true if the entry has been evicted from the cache, otherwise false
	 * @throws PersistenceException if the instance is dirty or used by other
	 * working set.
	 */
	boolean evict(WorkingSet context, Object oid, boolean force)
			throws PersistenceException;

	/**
	 * This methods permits to try the cache eviction of all persistent object
	 * @return the number of entry staying in the cache (not evicted)
	 */
	int evictAll(WorkingSet context, boolean force) throws PersistenceException;

	/**
	 * Detach a persistent object from the persistent support. After calling
	 * this method the persistent object is no more associated to the context
	 * and this instance is no more in the cache. If the persistent object has
	 * been modified, it is flushed. Be carefull the lock for the concurrency
	 * (read | write) is kept until the end of the context.
	 *
	 * @param ws is the current working set
	 * @param oid is the identifier of the persistent object
	 * @throws PersistenceException
	 */
	void unbind(WorkingSet ws, Object oid) throws PersistenceException;

	/**
	 * Detach a persistent object from the persistent support. After calling
	 * this method the persistent object is no more associated to the context
	 * and this instance is no more in the cache. If the persistent object has
	 * been modified, it is flushed. Be carefull the lock for the concurrency
	 * (read | write) is kept until the end of the context.
	 *
	 * @param ws is the current working set
	 * @param ce is the cache entry of the persistent object
	 * @throws PersistenceException
	 */
	void unbind(WorkingSet ws, CacheEntry ce) throws PersistenceException;

	/**
	 * Forget modifications of a dirty persistent object. The state is re-read
	 * from the cache or later from the database. Be carefull the lock for
	 * the concurrency (read | write) is kept until the end of the context.
	 *
	 * @param ws is the current working set
	 * @param oid is the identifier of the persistent object
	 * @throws PersistenceException
	 */
	void refresh(WorkingSet ws, Object oid) throws PersistenceException;

	/**
	 * Forget modifications of a dirty persistent object. The state is re-read
	 * from the cache or later from the database. Be carefull the lock for
	 * the concurrency (read | write) is kept until the end of the context.
	 *
	 * @param ws is the current working set
	 * @param ce is the cache entry of the persistent object
	 * @throws PersistenceException
	 */
	void refresh(WorkingSet ws, CacheEntry ce) throws PersistenceException;

	/**
	 * creates a working set
	 * @param userObject can be an external user depending on the
	 * personnality.
	 * @return a new WorkingSet
	 * @throws PersistenceException
	 */
	WorkingSet createWS(Object userObject)
			throws PersistenceException;

	/**
	 * creates a working set
	 * @param userObject can be an external user depending on the
	 * personnality.
	 * @param workingSetType is a type of the working set.
	 * @return a new WorkingSet
	 * @throws PersistenceException
	 */
	WorkingSet createWS(Object userObject,
						Object workingSetType)
			throws PersistenceException;

	/**
	 * Closes and releases a context.
	 * @param context which must be closed
	 * @throws PersistenceException if the context is a transactional context
	 * which has not been commited or aborted.
	 */
	void close(WorkingSet context) throws PersistenceException;

}
