/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.persistence.api;

import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheEventListener;

/**
 * A StateManager is used by the PersistenceManager and the ConcurrencyManager
 * in order to manage the life cycle of cache entries and their states. This
 * management concerns the java life cycle of the state instance (create,
 * destroy) and the status (dirty, new, deleted, ...).  
 * This interface is implemented by the Personality using the persistence
 * framework of Perseus.
 * @author S.Chassande-Barrioz
 */
public interface StateManager extends CacheEventListener {

	/**
	 * Creates a new State.
	 * @return a State instance linked to its cache entry (ie the
	 * state.getCacheEntry() must return 'ce')
	 */
	State createState(CacheEntry ce);

	/**
	 * Creates a new State.
	 * @param s the state to clone
	 * @return a new state which contains the same values (data and cache entry
	 * reference) than the existing state given in parameter
	 */
	State createState(State s);

	/**
	 * @param ce is the cache entry when the reference state is asked
	 * @return the reference state of a cache entry. If no reference state
	 * exists, the 'null' value is retrieved.
	 */
	State getReferenceState(CacheEntry ce);

	/**
	 * Assignes the reference state to a cache entry
	 * @param ce is the cache entry which the reference state must be changed
	 * @param state is the new value of the reference state. If this parameter
	 * is null, that means the CacheEntry does not have a state in memory, it
	 * is hollow.
	 */
	void setReferenceState(CacheEntry ce, State state);

	/**
	 * Removes a state. In general this method does nothing, but it can be used
	 * as a listener.
	 * @param state the state to destroy
	 */
	void destroyState(State state);

	/**
	 * Marks the status of the state to 'unexported' (deleted).
	 */
	void makeUnexported(State state);

	/**
	 * @return true if the state has been marked as unexported, otherwise false
	 * The state is unexported if the makeUnexported method has been called.
	 */
	boolean isUnexported(State state);

	/**
	 * Marks the status of the state to 'exported' (created).
	 */
	void makeExported(State state);

	/**
	 * @return true if the state has been marked as exported, otherwise false
	 * The state is exported if the makeExported method has been called.
	 */
	boolean isExported(State state);

	/**
	 * Marks the status of the state to 'dirty' (modified). The next calls to
	 * isflushed will return false.
	 */
	void makeDirty(State state);

	/**
	 * @return true if the state has been marked as dirty, otherwise false
	 * The state is dirty if the makeDirty or makeExported or makeUnexported
	 * method has been called.
	 */
	boolean isDirty(State state);

	/**
	 * Marks the status of the state to 'clean' (non modified). This method is
	 * used at the commit of a working set in order to reset the state status.
	 */
	void makeClean(State state);

	/**
	 * Marks the status of the state to 'flushed' (written on data support).
	 * @param state
	 */
	void makeFlushed(State state);

	/**
	 * @return true if the state has been marked as flushed without call to
	 * makeDirty, otherwise false
	 */
	boolean isFlushed(State state);

	/**
	 * A persistent object has been removed or evicted from the cache, then
	 * it must be marked as non persistent
	 * @param ce is the cache entry referenced the persistent object to unbind
	 */
	void makeUnbound(CacheEntry ce);

	/**
	 * A persistent object has been removed or evicted from the cache, then
	 * it must be marked as non persistent
	 * @param ce is the cache entry referenced the persistent object to bind
	 * @param oid is the id of the cache entry to bind 
	 */
	void makeBound(CacheEntry ce, Object oid);
	
	/**
	 * @return true if the persistent object referenced by the given cache
	 * entry, is persistent.
	 */
	boolean isBound(CacheEntry ce);
	
    /**
     * Indicates if a State has to be merged at commit time with the reference
     * state
     */
	boolean isToMerge(State state);
    
    /**
     * Marks the flag 'toMerge' on a state. a true value indicates that the
     * state has to be merged at commit time, with the reference state. 
     * @param state
     * @param thinLock is a hints describing  the thinlock causing the merge
     * at commit time.
     */
	void makeToMerge(State state, Object thinLock);
    
    /**
     * Merges a state (newState) into a oldSate. 
     * @param oldState is the state wich will contains the delta of the newState
     * @param newState contains the delta to merge to the oldState
     * @returnis the result of the merge operation  
     */
	State merge(State oldState, State newState);

	/**
	 * Indicates that this state is no more used by any working set. This event
	 * permit for instance to unswizzle the references to persistent objects 
	 * from the state.
	 * @param state is the state that is not used by any working set
	 */
	void stateNoMoreUsed(State state);
	
	/**
	 * The states of the persistent objects must not be kept in the cache
	 */
	int CACHING_STRATEGY_NO_CACHE 	= 1;
	
	/**
	 * The states of the persistent objects must be kept in the cache
	 */
	int CACHING_STRATEGY_CACHED 	= 2;

	/**
	 * The states of the persistent objects must be kept in the cache. In 
	 * addition, the persistent objects are fixed in the cache in order to 
	 * avoid its automatic eviction.
	 */
	int CACHING_STRATEGY_FIXED 		= 3;

	/**
	 * The states of the persistent objects must be kept in the cache. In 
	 * addition, <b>all</b> persistent objects of a class are present and fixed
	 * in the cache in order to avoid their automatic eviction.
	 */
	int CACHING_STRATEGY_ALL 		= 4;
	
	/**
	 * Indicates the caching strategy for a persistent class. The possible value
	 * are the following: 
	 * - CACHING_STRATEGY_NO_CACHE,
	 * - CACHING_STRATEGY_CACHED 
	 * - CACHING_STRATEGY_FIXED 
	 * - CACHING_STRATEGY_ALL
	 *  
	 * @see #CACHING_STRATEGY_NO_CACHE
	 * @see #CACHING_STRATEGY_CACHED
	 * @see #CACHING_STRATEGY_FIXED
	 * @see #CACHING_STRATEGY_ALL
	 * @param o is a persistent object or its identifier
	 */
	int getCachingStrategy(Object o);
}
