/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */
package org.objectweb.perseus.persistence.api;

import java.util.Map;
import java.util.Set;

/**
 * It defines a context used in by the PersistenceManager. A context contains
 * a collection of CacheEntry instances and has a status.
 *
 * @author S.Chassande-Barrioz, Y.Bersihand
 */
public interface WorkingSet {

	/**
	 * The status of a non transactional context which is not already
	 * closed.
	 */
	byte CTX_ACTIVE = 1;


	/**
	 * The status of a cloased context (transactional or not)
	 */
	byte CTX_CLOSED = 32;

	/**
	 * The origin of the binding action.
	 */
	byte READ_INTENTION = 1;
	byte WRITE_INTENTION = 2;
	byte UNKNOWN_INTENTION = 3;
	
	/**
	 * Returns the working set's status.
	 * @return one of the constants of the {@link
	 * org.objectweb.perseus.persistence.api.TransactionalPersistenceManager}
	 * interface.
	 */
	byte getStatus();

	/**
	 * Sets the working set's status.
	 * This setter permits also to intercept the working set life cycle.
	 * @param status one of the constants of the {@link
	 * org.objectweb.perseus.persistence.api.TransactionalPersistenceManager}
	 * interface.
	 * @throws PersistenceException if this is a probleme in the interception of
	 * the working set life cycle.
	 */
	void setStatus(byte status) throws PersistenceException;

	/**
	 * @return the entry attached to the working set weither its identifier
	 */
	State lookup(Object oid);

	/**
	 * Attaches an entry to the working set.
	 * @param state is the state which must be attached to the working set
	 * @param oid is the id of the corresponding object
	 * @param mode is the origin of the binding action: read/write intention 
	 * @see READ_INTENTION.WRITE_INTENTION.UNKNOWN_INTENTION
	 * @return the value of the old state in the cache, can be null if the state was not already in the cache
	 */
	
	State bind(State state, Object oid, byte mode);

	/**
	 * Dettaches an entry from the working set.
	 * @param oid is the identifier of entry which must be detattached from
	 * the working set
     * @return true if the entry was present in the working set
	 */
	boolean unbind(Object oid);

	/**
	 * Removes all entries of the working set.
	 */
	void clear();

	/**
	 * @return the list of entries attached to the working set.
	 */
	Set entries();

    /**
	 * @return the list of oids attached to the working set
	 */
	Set oids();

	/**
	 * 
	 * @return the map of (oid,entry) couples attached to the working set 
	 */
	Map getOid2Entry();
	
	/**
	 * @return the user object linked to the working set
	 */
	Object getUserObject();

	/**
	 * Compare the current working set to the parameter.
	 */
	//boolean equals(Object ws);

	/**
	 * @return a ConnectionHolder instance permiting to acces the data support
	 */
	ConnectionHolder getConnectionHolder();

	/**
	 * Indicates if the state of persistent objects reached in the working set
	 * must be kept in the cache at the working set end.
	 */
	boolean getWSRetainValues();

	/**
	 * assing a boolean a value Indicating if the state of persistent objects 
	 * reached in the working set must be kept in the cache at the working 
	 * set end.
	 */
	void setWSRetainValues(boolean val);

	/**
	 * Indicates if the dirty object must be reload at rollback time.
	 */
	boolean getWSRestoreValues();

	/**
	 * assignes a boolean value indicating if the dirty object must be reload 
	 * at rollback time.
	 */
	void setWSRestoreValues(boolean val);
	
	/**
	 * Indicates if the persistent object used in this working set can be shared
	 * with another transaction. If false that means java persistant instance 
	 * are be used only by this working set. 
	 */
	boolean getWSObjectSharing();
	
	/**
	 * Assignes a boolean value indicating if the persistent object used in 
	 * this working set can be shared with another transaction. If false that 
	 * means java persistant instance are be used only by this working set. 
	 */
	void setWSObjectSharing(boolean v);
}
