/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.persistence.concurrency;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.StateManager;
import org.objectweb.perseus.persistence.api.StorageManager;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is an implementation of a PConcurrencyManager which delegates the
 * concurrency control to the database.
 * 
 * @author P. Dechamboux
 * @author S.Chassande-Barrioz
 */
public class PDbDelegateConcurrencyManager
		implements PConcurrencyManager, BindingController, LifeCycleController {

	private String componentState = LifeCycleController.STOPPED;

    protected Logger logger = null;

    protected StateManager stateManager;
    protected StorageManager storageManager;

	//IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
        return new String[] {
        		STORAGE_MANAGER_BINDING,
        		STATE_MANAGER_BINDING,
        };
    }

    public Object lookupFc(String s) throws NoSuchInterfaceException {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            return storageManager;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            return stateManager;
        } else {
            throw new NoSuchInterfaceException(s);
        }
    }

    public void bindFc(String s, Object o) 
    	throws NoSuchInterfaceException, IllegalBindingException {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = (StorageManager) o;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = (StateManager) o;
        } else {
            throw new NoSuchInterfaceException(s);
        }
    }

    public void unbindFc(String s) throws NoSuchInterfaceException {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = null;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = null;
        } else {
            throw new NoSuchInterfaceException(s);
        }
    }

    public void finalize(Object ctx) {
	}
	//IMPLEMENTATION OF THE LifeCycleController INTERFACE //
	//----------------------------------------------------//
	public String getFcState() {
		return componentState;
	}

	public void startFc() throws IllegalLifeCycleException {
		if (componentState == LifeCycleController.STARTED) return;
		componentState = LifeCycleController.STARTED;
	}

	public void stopFc() throws IllegalLifeCycleException {
		if (componentState == LifeCycleController.STOPPED) return;
		componentState = LifeCycleController.STOPPED;
	}

	//IMPLEMENTATION OF THE ConcurrencyManager INTERFACE //
	//---------------------------------------------------//

	public void begin(Object ctx) {
	}

	public boolean validate(Object ctx) {
		return true;
	}

	public Object readIntention(Object ctx, Object resourceId, Object hints,
            ResourceProvider resProv, Object resourceHints)
			throws ConcurrencyException {
		return accessIntention(ctx, resourceId, hints, false, resProv, resourceHints);
	}

	public Object writeIntention(Object ctx, Object resourceId, Object hints,
            ResourceProvider resProv, Object resourceHints)
			throws ConcurrencyException {
		return accessIntention(ctx, resourceId, hints, true, resProv, resourceHints);
	}

	public void abort(Object ctx) {
	}

	protected Object accessIntention(Object ctx,
							  Object resourceid,
							  Object thinLock,
							  boolean isWrite,
                              ResourceProvider resProv, 
                              Object resourceHints) throws ConcurrencyException {
		CacheEntry ce = (CacheEntry) resourceHints;
		WorkingSet ws = (WorkingSet) ctx;
		if (resProv == null) {
			//export, or unexport(OID)
			return null;
		} else {
			return load(ws, resourceid, ce, isWrite);
		}
	}

	private State load(WorkingSet ws, Object resourceId, CacheEntry ce, 
			boolean isWrite) throws ConcurrencyException {
		State state = stateManager.createState(ce);
		try {
			storageManager.read(ws, resourceId, state, isWrite);
			stateManager.makeClean(state);
		} catch (PersistenceException e) {
			stateManager.destroyState(state);
			throw new NoDSIConcurrencyException(e);
		}
		return state;
	}
}
