/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.persistence.lib;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheEntryFactory;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.CacheFullException;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.perseus.cache.api.FixableCacheEntry;
import org.objectweb.perseus.cache.api.UnbindManager;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.concurrency.api.RolledBackConcurrencyException;
import org.objectweb.perseus.concurrency.pessimistic.Lock;
import org.objectweb.perseus.persistence.api.MemoryInstanceManager;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.RolledBackPersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.StateFilter;
import org.objectweb.perseus.persistence.api.StateManager;
import org.objectweb.perseus.persistence.api.StorageManager;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManagerAttributeController;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSetFilter;
import org.objectweb.perseus.persistence.api.WorkingSetLifeCycle;
import org.objectweb.perseus.persistence.api.WorkingSetManager;
import org.objectweb.perseus.persistence.api.VirtualState;
import org.objectweb.perseus.persistence.api.PersistentObjectLifeCycleExcpetion;
import org.objectweb.perseus.persistence.api.NoDSIPersistenceException;
import org.objectweb.perseus.persistence.concurrency.NoDSIConcurrencyException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.ArrayList;

/**
 * This class is an implementation of the TransactionalPersistenceManager
 * interface. It manages the persistency, the concurrency and the caching of
 * persistent objects. To do this it uses several other components:
 * - a cache manager,
 * - a concurrency manager,
 * - a storage manager,
 * - a state manager,
 * - a memory instance manager,
 * - a working set manager.
 * This implementation is compliant to fractal by implementing the
 * BindingController interface permitting the management of bindings between
 * components.
 *
 * @author S.Chassande-Barrioz, P.Dechamboux, O.Lobry, Y. Bersihand
 */
public class TransactionalPersistenceManagerImpl
		implements TransactionalPersistenceManager, BindingController,
		TransactionalPersistenceManagerAttributeController {

	public final static String CACHE_MANAGER_BINDING = "cache-manager";
	public final static String CONCURRENCY_MANAGER_BINDING = "concurrency-manager";
	public final static String STORAGE_MANAGER_BINDING = "storage-manager";
	public final static String STATE_MANAGER_BINDING = "state-manager";
	public final static String MEMORY_INSTANCE_MANAGER_BINDING = "memory-instance-manager";
	public final static String UNBIND_MANAGER_BINDING = "unbind-manager";
	public final static String WORKINGSET_MANAGER_BINDING = "workingset-manager";
	public final static String CACHE_ENTRY_FACTORY_BINDING = "cache-entry-factory";

	/**
	 * The cache of the persistent object.
	 */
	protected CacheManager cache;

	protected UnbindManager unbindManager;

	/**
	 * The concurrency manager
	 */
	protected ConcurrencyManager cm;

	/**
	 * The storage manager in charge of the I/O to the data support, and the
	 * naming management.
	 */
	protected StorageManager storage;

	/**
	 * The manager of state of persistent object.
	 */
	protected StateManager stateManager;

	/**
	 * The manager of working set instance.
	 */
	protected WorkingSetManager wsManager;

	/**
	 * The factory of persistent objects
	 */
	protected MemoryInstanceManager mim;

	protected Logger logger = null;

	/**
	 * The list of filters to order the working sets
	 */
	protected Collection filters = null;
	
	protected CacheEntryFactory cef = null;
	
	/**
	 * @deprecated 
	 * Attaches a state to a working Set. 
	 *
	 * @param state is the state of the persistent object which must be attached
	 * to working set.
	 * @param ws is the working set to attach the state.
	 */
	protected void bindCeInWS(State state, WorkingSet ws, byte mode) {
		Object oid = state.getCacheEntry().getCeIdentifier();
		ws.bind(state, oid, mode);
	}


	/**
	 * finds or creates the CacheEntry matching to an identifier. If no cache
	 * antry matching the identifier is found in the cache, an new instance
	 * is created without state.
	 * @param ws if the WorkingSet requiring a CacheEntry.
	 * @param oid the identifier of the expected cache entry
	 * @return a cache entry instance (never null).
	 * @throws PersistenceException if it is not possible to instanciate a
	 * persistent object
	 * @throws RolledBackPersistenceException if the cache is full (case of the
	 * entry has not been found into the cache and a new entry is bound).
	 */
	protected CacheEntry getCacheEntry(WorkingSet ws,
									   Object oid) throws PersistenceException {
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "getCacheEntry( " + oid + ")");
		}
		//Check in the working set if the persistent object is not already used
		State state = ws.lookup(oid);
		CacheEntry ce;
		if (state != null && state != VirtualState.instance) {
			ce = state.getCacheEntry();
		} else if (isShareableObject(oid, ws)) {
			// the cache is used at least for sharing the same persistent java 
			// object. According to the caching strategy the cache can be used 
			// to keep state (persistent data) in addition
			ce = cache.lookup(oid);
			if (ce == null) {
				Object po = mim.newInstance(oid, ws.getConnectionHolder());
				// add memory instance to cache
				try {
					ce = cache.bind(oid, po);
				} catch (CacheFullException e) {
					throw new RolledBackPersistenceException(e);
				} catch (CacheException e) {
					//That means the entry has been already bind to the cache. 
					// There are concurrent accesses.
					ce = cache.lookup(oid);
					if (ce == null) {
					    logger.log(BasicLevel.WARN, 
					            "Recursive call for getting a cache entry (oid= " 
					            + oid + ")");
					    ce = getCacheEntry(ws, oid);
					}
				}
			}
		} else { //no object sharing
			//create a new persistent object
			Object po = mim.newInstance(oid, ws.getConnectionHolder());
			//create the cache entry (not an entry in the cache but only
			// a java object implementing the CacheEntry interface)
			ce = cef.create(oid, po);
			//Remove the reference in order to force the data loading
			stateManager.setReferenceState(ce, null);
		}
		return ce;
	}
	
	/**
	 * Check if a cache entry is already referenced in the cache. If the entry
	 * has not been found, the entry is bound in the cache.
	 *
	 * @param ce is the cache entry to check
	 * @throws PersistenceException if the cache is full or if another cache
	 * entry has been found in the cache with the same identifier.
	 */
	protected void checkCacheEntry(CacheEntry ce, WorkingSet ws) throws PersistenceException {
		Object oid = ce.getCeIdentifier();
		if (oid == null) {
			throw new NoDSIPersistenceException(
					PersistentObjectLifeCycleExcpetion.READ_NPO_MSG);
		}
		if (!stateManager.isBound(ce)) {
		    //The cache entry has been unbound. 
		    unbindNonPersistentInstance(ce.getCeIdentifier(), ws);
		    throw new NoDSIPersistenceException(
					PersistentObjectLifeCycleExcpetion.READ_NPO_MSG);
		}
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "checkCacheEntry " + oid);
		}
		State state = ws.lookup(oid);
		Object po = ce.getCeObject();
		if (state != null && state != VirtualState.instance) {
			// the persistent object already has a real state bound to the 
			// working set. So we can suppose the checking has been done the 
			// first time
			return;
		} else if (isCacheable(po, ws)) {
			CacheEntry entryOfCache = cache.lookup(oid);
			if (entryOfCache == null) {
				//the entry is not present in the cache
				try {
					entryOfCache = cache.bind(oid, ce.getCeObject());
				} catch (CacheFullException e) {
					throw new RolledBackPersistenceException(e);
				} catch (CacheException e) {
					//That means the entry has been already bind to the cache. 
					// There are concurrent accesses.
					if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
					    logger.log(BasicLevel.DEBUG, "checkCacheEntry " + oid);
					}
				}
			} else if (ce != entryOfCache && isShareableObject(po, ws)) {
				//the entry present in the cache is not the same than the parameter
				throw new RolledBackPersistenceException(
						"Another persistent object is already present into the cache (identifier=" + oid + ")");
			}
		}
	}
	
	/**
	 * Indicates if a persistent object could have a reference state in the
	 * cache.
	 * @param o a persistent object or its identifier
	 * @return true is a reference state corresponding to the persistent object
	 * could be found in cache, false otherwise. 
	 */
	private boolean isStateCacheable(Object o) {
		return stateManager.getCachingStrategy(o) > StateManager.CACHING_STRATEGY_NO_CACHE;
	}
	
	/**
	 * Indicates if a persistent object used in a working set is shareable 
	 * with other working sets
	 * @param o a persistent object or its identifier
	 * @param ws the working set asking the question and using the persistent
	 * object
	 * @return
	 */
	private boolean isShareableObject(Object o, WorkingSet ws) {
		return ws.getWSObjectSharing() && mim.isObjectSharing(o);
	}

	/**
	 * Indicates if a persistent object could be the cache according to the 
	 * configuration.
	 * @param o a persistent object or its identifier
	 * @param ws the working set asking the question and using the persistent
	 * object
	 * @return true is the persistent object could be found in cache, false
	 * otherwise. 
	 */
	private boolean isCacheable(Object o, WorkingSet ws) {
		return isStateCacheable(o) || isShareableObject(o, ws);
	}
	

	/**
	 * Prepares the end of a working set.
	 * @param ws is can be simple working set or a transaction
	 * @return true if the context CAN be committed, false if the context MUST be
	 * rolled back
	 * @throws PersistenceException
	 */
	private boolean prepareWS(WorkingSet ws, boolean validate)
			throws PersistenceException {
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "prepareWS " + ws + " / " + validate);
		}
		boolean valid = validate;
		PersistenceException pe = null;
		try {
			ws.setStatus(WorkingSetLifeCycle.getNextStatus(
					ws.getStatus(), WorkingSetLifeCycle.PREPARE_ACTION));
		} catch (PersistenceException e) {
			pe = e;
			valid = false;
		}
		try {
			// check if the context can be validate on concurrency manager
			valid = cm.validate(ws) && valid;
			if (pe != null) {
				throw pe;
			}
			if (valid) {
				//order the ws according to the filter(s)
				if (filters != null) {
					Iterator it = filters.iterator();
					while (it.hasNext()) {
						WorkingSetFilter wsf = (WorkingSetFilter) it.next();
						wsf.order(ws);
					}
				}
				//flush dirty objects into data support,
				Set entries = ws.entries();
				for (Iterator it = entries.iterator(); it.hasNext();) {
					State state = (State) it.next();
					if (state == VirtualState.instance) {
						continue;
					}
					if (stateManager.isDirty(state) 
							&& !stateManager.isFlushed(state)) {
						storage.write(ws.getConnectionHolder(),
								state.getCacheEntry().getCeIdentifier(),
								state);
					}
				}
			}
		} catch(Exception e) {
			// If an exception occurs then the working set must be cancelled
			valid = false;
			throw (e instanceof PersistenceException
				? (PersistenceException) e
				: new PersistenceException("An exception occurs during the "
					+ "working preparation, it is marked as rolled back",e));
		} finally {
			//release the data support connection
			ws.getConnectionHolder().releaseCHConnection();
			if (valid) {
				ws.setStatus(WorkingSetLifeCycle.getNextStatus(
						ws.getStatus(), WorkingSetLifeCycle.PREPARE_OK_ACTION));
			} else {
				ws.setStatus(WorkingSetLifeCycle.getNextStatus(
						ws.getStatus(), WorkingSetLifeCycle.PREPARE_FAIL_ACTION));
			}
		}
		return valid;
	}

	/**
	 * commit a context.
	 * @param ws is can be simple working set or a transaction
	 * @param action is the action declenching the rollback
	 * @throws PersistenceException
	 */
	private void commitWS(WorkingSet ws,
						  byte action) throws PersistenceException {
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "commitWS " + ws + " / " + action);
		}
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				ws.getStatus(), action);
		try {
			ws.getConnectionHolder().commitCH();
		} catch (PersistenceException e) {
			rollbackWS(ws, WorkingSetLifeCycle.ROLLBACK_ACTION);
			throw new RolledBackPersistenceException(e);
		}
        boolean retainValues = ws.getWSRetainValues();

        try {
			ArrayList unexportedOids = null;
			for (Iterator it = ws.entries().iterator(); it.hasNext();) {
				State state = (State) it.next();
				if (state == VirtualState.instance 
                        || !stateManager.isDirty(state)) {
					continue;
				}
                CacheEntry ce = state.getCacheEntry();
				if (stateManager.isUnexported(state)) {
                    if (isCacheable(ce.getCeObject(), ws)) {
                        //Register the entry to unexport
                        if (unexportedOids == null) {
    						unexportedOids = new ArrayList();
    					}
                        unexportedOids.add(ce.getCeIdentifier());
                    }
                    if (stateManager.getReferenceState(ce) != state) {
                        stateManager.setReferenceState(ce, state);
                    }
					//Unbind the entry from the persistent layer
					stateManager.makeUnbound(ce);
				} else {
                    Object po = ce.getCeIdentifier();
                    boolean shareable = isShareableObject(po, ws); 
                    boolean cacheable = isStateCacheable(po);
                    if (retainValues && cacheable) {
                        final Object oid = ce.getCeIdentifier();
                        CacheEntry entryOfCache;;
                        if (!shareable) {
                            //when java persistent object is not shareable
                            // the cache entry used in the working set is not
                            //the one of the cache
                            entryOfCache = cache.lookup(oid);
                            //be careful the real cache entry could be already 
                            // evicted
                            if (entryOfCache == null) {
                                stateManager.destroyState(state);
                                continue;
                            }
                        } else {
                            entryOfCache = ce;
                        }
                        if (stateManager.isToMerge(state)) {
                            //A thin lock has been allocated, then several
                            //working set potentialy commit the state delta.
                            //==>synchronize the state change
                            synchronized(entryOfCache.getCeIdentifier()) {
                                State s = stateManager.getReferenceState(entryOfCache);
                                if (s != null) {
                                    s = stateManager.merge(s, state);
                                    stateManager.setReferenceState(entryOfCache, s);
                                }
                            }
                        } else {
                            // a single writer is autorized 
                            // ==> no synchronization required
                            stateManager.makeClean(state);
                            State rs = stateManager.getReferenceState(entryOfCache);
                            // if there is not reference state before it means 
                            // it has been removed between the tx. Perhaps due 
                            // to an evictAll call on the L2 cache. then the 
                            // reference state is not set here in order to 
                            // reload data next time.
                            if (rs != null && rs != state) {
                                stateManager.setReferenceState(entryOfCache, state);
                            }
                        }
                    } else if (stateManager.getReferenceState(ce) == state) {
                        //state is clean or no caching or no retains value
                        stateManager.makeClean(state);
                        //forget the state used as reference state
                        stateManager.setReferenceState(ce, null);
                        //destroy the state
                        stateManager.destroyState(state);
                    }                    
                }
			}
            try {
    			if (unexportedOids != null) {
    				// unbind removed entries from the cache
    				unbindManager.unbindAll(unexportedOids, true);
    			}
    		} catch (CacheException e) {
    			throw new PersistenceException(e);
            }
		} finally {
			try {
				ws.getConnectionHolder().closeCHConnection();
			} finally {
				storage.endWS(ws);
				cm.finalize(ws);
				ws.setStatus(newStatus);
				ws.clear();
			}
		}
	}

	/**
	 * cancel a context. The unfix action is out of date.
	 * @param ws is can be simple working set or a transaction
	 * @param action is the action declenching the rollback
	 * @throws PersistenceException
	 */
	private void rollbackWS(WorkingSet ws,
							byte action) throws PersistenceException {
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "rollbackWS " + ws + " / " + action);
		}
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				ws.getStatus(), action);
        try {
            ws.getConnectionHolder().rollbackCH();
        } finally {
            ws.setStatus(newStatus);
        }
        try {
			ArrayList unexportedOids = null;
			for (Iterator i = ws.entries().iterator(); i.hasNext();) {
				State state = (State) i.next();
				if (state == VirtualState.instance 
                        || !stateManager.isDirty(state)) {
					continue;
				}
				CacheEntry ce = state.getCacheEntry();
				if (stateManager.isExported(state)) {
					if (unexportedOids == null) {
						unexportedOids = new ArrayList();
					}
					unexportedOids.add(ce.getCeIdentifier());
                    stateManager.makeClean(state);
					stateManager.makeUnbound(ce);
                } else {
                    if (stateManager.isUnexported(state)) {
                        //set the persistent object in a right state even if the
                        // state will be destroyed below
                        stateManager.makeClean(state);
                        stateManager.makeBound(ce, ce.getCeIdentifier());
                    }
                    stateManager.setReferenceState(ce, null);
                    //destroy the state
                    stateManager.destroyState(state);
                }
			}
			if (unexportedOids != null) {
				// unbind removed entries from the cache
                try {
                    unbindManager.unbindAll(unexportedOids, true);
                } catch (CacheException ce) {
                    throw new PersistenceException("Impossible to rollback - ", ce);
                }                    
			}
		} finally {
			try {
				ws.getConnectionHolder().closeCHConnection();
			} finally {
				storage.endWS(ws);
				cm.abort(ws);
				ws.clear();
				ws.setStatus(newStatus);
			}
		}
	}

	private void assertOpenWorkingset(WorkingSet ws) throws PersistenceException {
		if (ws.getStatus() == WorkingSet.CTX_CLOSED) {
			throw new PersistenceException("Impossible to work with a closed WorkingSet");
		}
	}

	//IMPLEMENTATION OF THE UserBindingControler INTERFACE //
	//------------------------------------------------------//

	public String[] listFc() {
		return new String[]{
			CACHE_MANAGER_BINDING,
			CONCURRENCY_MANAGER_BINDING,
			STATE_MANAGER_BINDING,
			STORAGE_MANAGER_BINDING,
			MEMORY_INSTANCE_MANAGER_BINDING,
			UNBIND_MANAGER_BINDING,
			WORKINGSET_MANAGER_BINDING,
			STATE_MANAGER_BINDING,
			CACHE_ENTRY_FACTORY_BINDING
		};
	}

	public Object lookupFc(String s) {
		if (CACHE_MANAGER_BINDING.equals(s)) {
			return cache;
		} else if (CONCURRENCY_MANAGER_BINDING.equals(s)) {
			return cm;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			return stateManager;
		} else if (STORAGE_MANAGER_BINDING.equals(s)) {
			return storage;
		} else if (MEMORY_INSTANCE_MANAGER_BINDING.equals(s)) {
			return mim;
		} else if (UNBIND_MANAGER_BINDING.equals(s)) {
			return unbindManager;
		} else if (WORKINGSET_MANAGER_BINDING.equals(s)) {
			return wsManager;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			return stateManager;
		} else if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
			return cef;
		}
		return null;
	}

	public void bindFc(String s, Object o) {
		if ("logger".equals(s)) {
			logger = (Logger) o;
		} else if (CACHE_MANAGER_BINDING.equals(s)) {
			cache = (CacheManager) o;
		} else if (CONCURRENCY_MANAGER_BINDING.equals(s)) {
			cm = (ConcurrencyManager) o;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = (StateManager) o;
		} else if (STORAGE_MANAGER_BINDING.equals(s)) {
			storage = (StorageManager) o;
		} else if (MEMORY_INSTANCE_MANAGER_BINDING.equals(s)) {
			mim = (MemoryInstanceManager) o;
		} else if (UNBIND_MANAGER_BINDING.equals(s)) {
			unbindManager = (UnbindManager) o;
		} else if (WORKINGSET_MANAGER_BINDING.equals(s)) {
			wsManager = (WorkingSetManager) o;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = (StateManager) o;
		} else if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
			cef = (CacheEntryFactory) o;
		}
	}

	public void unbindFc(String s) {
		if (CACHE_MANAGER_BINDING.equals(s)) {
			cache = null;
		} else if (CONCURRENCY_MANAGER_BINDING.equals(s)) {
			cm = null;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = null;
		} else if (STORAGE_MANAGER_BINDING.equals(s)) {
			storage = null;
		} else if (MEMORY_INSTANCE_MANAGER_BINDING.equals(s)) {
			mim = null;
		} else if (UNBIND_MANAGER_BINDING.equals(s)) {
			unbindManager = null;
		} else if (WORKINGSET_MANAGER_BINDING.equals(s)) {
			wsManager = null;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = null;
		} else if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
			cef = null;
		}
	}

    //IMPLEMENTATION OF THE ConcurrencyManager.ResourceProvider INTERFACE //
    //--------------------------------------------------------------------//
    public Object getResource(Object _ws, 
            Object resourceId, 
            Object resourceHints,
            Object lockHints, 
            boolean resourceCopyRequired, 
            Object suggestedResource, 
            Object synchronizer) throws ConcurrencyException {
        if (suggestedResource != null) {
            return suggestedResource;
        }
        WorkingSet ws = (WorkingSet) _ws;
        State state = null;
        CacheEntry ce = (CacheEntry) resourceHints;
        Object po = ce.getCeObject();
        // new persistent object from the working set
        if (isStateCacheable(po)) {
            // get the reference state from the cache
            if (isShareableObject(po, ws)) {
                // java instance are shared, then there is only one cache
                // entry.
                state = stateManager.getReferenceState(ce);
                if (state == null) { // no reference state ==> load it
                    // avoid multiple loading of the same persistent object
                    synchronized(synchronizer) {
                        state = stateManager.getReferenceState(ce);
                        if (state == null) {
                            state = load(ws, ce);
                            //put in the cache
                            stateManager.setReferenceState(ce, state);
                        }
                    }
                }
                if (resourceCopyRequired || lockHints != null) {
                    // clone the reference state for this working set
                    state = stateManager.createState(state);
                    if (lockHints != null) {
                        stateManager.makeClean(state);
                        stateManager.makeToMerge(state, lockHints);
                    }
                }
            } else {
                // when the java reference is not shared, the cache entry
                // used in a working set is NOT the one from the cache,
                // which really contains the reference state.
                
                //Try to found a state from the cache (without synchronized block)
                final CacheEntry entryOfCache = cache.lookup(resourceId);
                if (entryOfCache != null) {
                    state = stateManager.getReferenceState(entryOfCache);
                }
                if (state == null) {
                    // no reference state in cache ==> we have to load it
                    synchronized(synchronizer) {
                        //check one more time in synchronized block
                        if (entryOfCache != null) {
                            state = stateManager.getReferenceState(entryOfCache);
                        }
                        if (state == null) {
                            state = load(ws, entryOfCache);
                        }
                    }
                    if (entryOfCache != null) {
                        // put the state in the cache
                        stateManager.setReferenceState(entryOfCache, state);
                    }
                } //else state has been found from the cache
                
                // ALWAYS clone found state and use it as reference state for the 
                // cache entry used in the working set 
                state = stateManager.createState(state);
                stateManager.makeClean(state);
                stateManager.setReferenceState(ce, state);
                state.setCacheEntry(ce);
                if (lockHints != null) {
                    stateManager.makeToMerge(state, lockHints);
                }
            }
        } else {
            //no state caching ==> load it
            state = load(ws, ce);
        }
        return state;
    }
    private State load(WorkingSet ws, CacheEntry ce) throws ConcurrencyException {
        State state = stateManager.createState(ce);
        Object oid = ce.getCeIdentifier();
        try {
            storage.read(ws, oid, state, false);
            stateManager.makeClean(state);
            return state;
        } catch (PersistenceException e) {
            stateManager.destroyState(state);
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Error during loading:" 
                        + "\n\t-oid=" + oid
                        + "\n\t-ws.lookup(oid)=" + ws.lookup(oid)
                        + "\n\t-ce.rs=" + stateManager.getReferenceState(ce)
                        + "\n\t-exception=" + e.getClass().getName() 
                        + ":" + e.getMessage());
            }
            if (e instanceof NoDSIPersistenceException) {
                throw new NoDSIConcurrencyException(e);
            } else {
                throw new ConcurrencyException(e);
            }
        }
    }
        
	// IMPLEMENTATION OF THE PersistenceManager INTERFACE //
	//----------------------------------------------------//

	public State export(WorkingSet ws, Object obj)
			throws PersistenceException {
		return export(ws, obj, null);
	}

	public State export(WorkingSet ws, Object obj, Object hints)
			throws PersistenceException {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "export ctx=" + ws
					+ " / obj=" + obj.getClass().getName()
					+ " / hints=" + hints);
		assertOpenWorkingset(ws);
		State state = null;
		//export through storage manager
		Object oid = null;
		CacheEntry ce = null;
		try {
            //Obtain an identifier
			if (hints == null) {
				oid = storage.export(ws.getConnectionHolder(), obj);
			} else {
				oid = storage.export(ws.getConnectionHolder(), obj, hints);
			}
            //Obtain a CacheEntry
            if (isCacheable(obj, ws)) {
                //add to cache
                ce = cache.bind(oid, obj);
            } else {
                //create the cache entry (not an entry in the cache but only
                // a java object implementing the CacheEntry interface)
                ce = cef.create(oid, obj);
            }
            //bind the current state to the working set
			state = stateManager.getReferenceState(ce);
            stateManager.makeExported(state);
            ws.bind(state, oid, WorkingSet.WRITE_INTENTION);
			//notify concurrency manager
			getWriteRight(ws, oid, null, state);
			return state;
		} catch (PersistenceException e) {
		    if (ce != null) {
			    stateManager.makeUnbound(ce);
		    }
			throw e;
		} catch (CacheFullException e) {
			throw new RolledBackPersistenceException(e);
		} catch (CacheException e) {
			//if the object has been deleted in the same tx, 
			//then its id can be re-used
			//virtual state???
			try {
				state = ws.lookup(oid);
				if (state != null && stateManager.isUnexported(state)) {
					ce = cache.lookup(oid);
					if (ce != null) {
						//	unbind removed entries from the cache
						unbindManager.unbind(oid, true);
						//add to cache
						ce = cache.bind(oid, obj);
						//bind the entry from the persistent layer
						//stateManager.makeBound(ce, oid);
                        state = stateManager.getReferenceState(ce);
                        if (state == null) {
                            state = VirtualState.instance;
                        }
						ws.bind(state, oid, WorkingSet.WRITE_INTENTION);
						//notify concurrency manager
						getWriteRight(ws, oid, null, state);
						//do not make it exported but make it clean then dirty 
                        stateManager.makeDirty(state);
                        stateManager.makeExported(state);
						return state;
					}
				}
			} catch (CacheException cacheE) {
				throw new PersistenceException(
				        "Cannot put object into cache, oid=" + oid, cacheE);
			}
			throw new PersistenceException(
			        "Cannot put object into cache, oid=" + oid, e);
		}
	}

	public State unexport(WorkingSet ws, Object oid) throws PersistenceException {
        assertOpenWorkingset(ws);
        CacheEntry ce = getCacheEntry(ws, oid);
        if (!stateManager.isBound(ce)) {
            throw new PersistentObjectLifeCycleExcpetion(
                    PersistentObjectLifeCycleExcpetion.UNEXPORT_NPO_MSG);
        }
        //Create a dummy state containing no data 
        State state = stateManager.createState(ce);
        //notify concurrency manager
        getWriteRight(ws, oid, null, state);
        //unexport through storage manager
        storage.unexport(ws.getConnectionHolder(), oid, ce);
        stateManager.makeUnexported(state);
        return state;
	}
	public State unexport(WorkingSet ws, CacheEntry ce) throws PersistenceException {
		assertOpenWorkingset(ws);
		if (!stateManager.isBound(ce)) {
			throw new PersistentObjectLifeCycleExcpetion(
					PersistentObjectLifeCycleExcpetion.UNEXPORT_NPO_MSG);
		}
        Object oid = ce.getCeIdentifier();
		//notify concurrency manager
        State state = getWriteRightNState(ws, ce, null);
		//unexport through storage manager
		storage.unexport(ws.getConnectionHolder(), oid, ce);
		stateManager.makeUnexported(state);
		return state;
	}

    public CacheEntry getObjectById(WorkingSet ws, Object oid)
    	throws PersistenceException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "getObjectById ctx=" + ws
                    + " / oid=" + oid);
        }
        assertOpenWorkingset(ws);
        //Check in the worling set if already associated
        State s = ws.lookup(oid);
        if (s != null && s != VirtualState.instance) {
            return s.getCacheEntry();
        }
        return getCacheEntry(ws, oid);
    }

	public State readIntention(WorkingSet ws, Object oid, Object thinLock)
		throws PersistenceException {
		assertOpenWorkingset(ws);
	    try {
            return _readIntention(ws, getCacheEntry(ws, oid), thinLock);
        } catch (NoDSIPersistenceException e) {
            unbindNonPersistentInstance(oid, ws);
            throw e;
        }
	}

	public State readIntention(WorkingSet ws, CacheEntry ce, Object thinLock) throws PersistenceException {
		assertOpenWorkingset(ws);
		checkCacheEntry(ce, ws);
        return _readIntention(ws, ce, thinLock);
	}

    /**
	 * This method records an access intention to a data object in read mode.
	 * @param ce is the cache entry
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @exception PersistenceException
	 */
	private State _readIntention(WorkingSet ws, CacheEntry ce, Object thinLock) throws PersistenceException {
        Object oid = ce.getCeIdentifier();
        State state = ws.lookup(oid);
        if (state != null) {
            if (state == VirtualState.instance) {
                state = null;
            } else {
                return state;
            }
        }
		// notify concurrency manager
		try {
			state = (State) cm.readIntention(ws, oid, thinLock, this, ce);
		} catch (RolledBackConcurrencyException e) {
			throw new RolledBackPersistenceException(e);
		} catch (NoDSIConcurrencyException e) {
		    unbindNonPersistentInstance(oid, ws);
			throw new NoDSIPersistenceException(e);
		} catch (ConcurrencyException e) {
			throw new PersistenceException(e);
		}
		ws.bind(state, oid, WorkingSet.READ_INTENTION);
		return state;
	}

	public State writeIntention(WorkingSet ws, Object oid, Object thinLock)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		try {
            return getWriteRightNState(ws, getCacheEntry(ws, oid), thinLock);
        } catch (NoDSIPersistenceException e) {
		    unbindNonPersistentInstance(oid, ws);
		    throw e;
        }
	}

	public State writeIntention(WorkingSet ws, CacheEntry ce, Object thinLock)
	throws PersistenceException {
	    assertOpenWorkingset(ws);
	    checkCacheEntry(ce, ws);
        return getWriteRightNState(ws, ce, thinLock);
	}
    public void getWriteRight(WorkingSet ws, 
            Object oid, 
            Object thinLock, 
            State state) throws PersistenceException {
        State wsstate = ws.lookup(oid);
        if (wsstate == state && stateManager.isDirty(state)) {
            return;
        }
        try {
            cm.writeIntention(ws, oid, thinLock, null, null);
        } catch (RolledBackConcurrencyException e) {
            throw new RolledBackPersistenceException(e);
        } catch (NoDSIConcurrencyException e) {
            unbindNonPersistentInstance(oid, ws);
            throw new NoDSIPersistenceException(e);
        } catch (ConcurrencyException e) {
            throw new PersistenceException(e);
        }
        ws.bind(state, oid, WorkingSet.WRITE_INTENTION);
        stateManager.makeDirty(state);
    }
    /**
     * Obtain the right to modify a persistent object. Obtain the state to use
     * in the working set
     * @param ws is the current working set asking the right
     * @param oid is the identifier of the persistent object
     * @param thinLock is an helper for concurrency permiting to reduce 
     * conflict. 
     * @return the state to use in the working set
     */
	public State getWriteRightNState(WorkingSet ws, 
            CacheEntry ce, 
            Object thinLock)
			throws PersistenceException {
        Object oid = ce.getCeIdentifier();
		State state = ws.lookup(oid);
        if (state != null) {
            if (state == VirtualState.instance) {
                state = null;
            } else if (stateManager.isDirty(state)) {
                return state;
            }
        }
		// notify concurrency manager
		try {
			state = (State) cm.writeIntention(ws, oid, thinLock, this, ce);
		} catch (RolledBackConcurrencyException e) {
			throw new RolledBackPersistenceException(e);
		} catch (NoDSIConcurrencyException e) {
		    unbindNonPersistentInstance(oid, ws);
			throw new NoDSIPersistenceException(e);
		} catch (ConcurrencyException e) {
			throw new PersistenceException(e);
		}
		ws.bind(state, oid, WorkingSet.WRITE_INTENTION);
		stateManager.makeDirty(state);
		return state;
	}

	private void unbindNonPersistentInstance(Object oid, WorkingSet ws) {
	    if (oid == null) {
	        return;
	    }
	    if (cache.lookup(oid) != null) {
			try {
				unbindManager.unbind(oid, true);
			} catch (CacheException e1) {
				logger.log(BasicLevel.ERROR,
						"Impossible to unbind from the cache an entry which " +
						"does not exist on the data support", e1);
			}
		}
		ws.unbind(oid);
	}
	/**
	 * This method can be used to mark the end of an access to a data object.
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param ce is the  cache entry on which the access is finished.
	 * @exception PersistenceException
	 */
	public void accessCompletion(WorkingSet ws, State ce) throws PersistenceException {
		assertOpenWorkingset(ws);
	}

	/**
	 * This method permits to flush entries associated to a context.
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param statefilter is a filter of the state which must be flushed
	 * @throws PersistenceException if the context is not valid.
	 */
	public void flush(WorkingSet ws, StateFilter statefilter)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		Iterator it = ws.entries().iterator();
		while (it.hasNext()) {
			State state = (State) it.next();
			if (state == VirtualState.instance) {
				continue;
			}
			if (stateManager.isDirty(state) 
					&& !stateManager.isFlushed(state)
			        && (statefilter == null || statefilter.accept(state))) {
				storage.write(ws.getConnectionHolder(),
						state.getCacheEntry().getCeIdentifier(),
						state);
				//Mark the entry as flushed in order to avoid I/O at prepare time
				stateManager.makeFlushed(state);
			}
		}
	}

	/**
	 * This method permits to flush entries associated to a context.
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param stateToFlush is the state taht must be flushed
	 * @throws PersistenceException if the context is not valid.
	 */
	public void flush(WorkingSet ws, State stateToFlush)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		State state = ws.lookup(stateToFlush.getCacheEntry().getCeIdentifier());
		if (state == stateToFlush 
				&& stateManager.isDirty(state) 
				&& !stateManager.isFlushed(state)) {
			storage.write(ws.getConnectionHolder(),
					state.getCacheEntry().getCeIdentifier(),
					state);
			//Mark the entry as flushed in order to avoid I/O at prepare time
			stateManager.makeFlushed(state);
		}
	}
	/**
	 * The unfix action is out of date.
	 */
	public boolean evict(WorkingSet ws, Object oid, boolean force) throws PersistenceException {
		FixableCacheEntry ce = (FixableCacheEntry) cache.lookup(oid);
		assertOpenWorkingset(ws);
		if (ce == null) {
			return true;
		}
		int fix = ce.getCeFixCount();
		State state = ws.lookup(oid);
		if (state != null) {
			//The entry is used only in the specified working set
			if (stateManager.isDirty(state) && !stateManager.isFlushed(state)) {
				//flush the dirty object and mark it as flushed
				storage.write(ws.getConnectionHolder(), oid, state);
				stateManager.makeFlushed(state);
			}
			//remove the entry from the working set by replaceing the state
			// by a virtual state.
			ws.bind(VirtualState.instance, oid, WorkingSet.UNKNOWN_INTENTION);
		}
		if (fix > 0) {
			if (force) {
				throw new PersistenceException("Impossible to evict an entry from the cache: There is another user than the given working set");
			} else {
				return false;
			}
		}
		//Remove the reference state
		stateManager.setReferenceState(ce, null);
		//Remove the local reference for the GC
		ce = null;
		//remove the entry from the cache
		try {
			return unbindManager.unbind(oid, false);
		} catch (CacheException e) {
			if (force) {
				throw new PersistenceException("Impossible to unbind the entry: ", e);
			} else {
				return false;
			}
		}
	}

	public int evictAll(WorkingSet ws, boolean force) throws PersistenceException{
		if (ws != null) {
		    assertOpenWorkingset(ws);
		}
		try {
			return unbindManager.unbindUnfixed(force).size();
		} catch (CacheException e) {
			throw new PersistenceException(e);
		}
	}

	public void unbind(WorkingSet ws, Object oid) throws PersistenceException {
		assertOpenWorkingset(ws);
		unbind(ws, getCacheEntry(ws, oid));
	}

	public void unbind(WorkingSet ws, CacheEntry ce) throws PersistenceException {
		assertOpenWorkingset(ws);
		FixableCacheEntry fce = (FixableCacheEntry) ce;
		
		Object oid = ce.getCeIdentifier();
		State state = ws.lookup(oid);
		if (state == null) {
		    state = readIntention(ws, ce, null);
		} else if (stateManager.isDirty(state) && !stateManager.isFlushed(state)) {
			//flush the dirty object and mark it as flushed
			storage.write(ws.getConnectionHolder(), oid, state);
			stateManager.makeFlushed(state);
		}
		// Remove the entry from the working set by replacing the real state
		// by a virtual state. It is required to keep the oid into the
		// working set in order to free the locking of the concurrency
		// manager
		ws.bind(VirtualState.instance, oid, WorkingSet.UNKNOWN_INTENTION);

        if (isCacheable(ce.getCeObject(), ws)) {
    		//remove the entry from the cache
    		try {
    			while (fce.getCeFixCount() > 0) {
    			    cache.unfix(fce);
    			}
    			unbindManager.unbind(oid, true);
    		} catch (CacheException e) {
    				throw new PersistenceException("Impossible to unbind the entry: ", e);
    		}
        }

		//Mark the cache entry as unbound from the Persistence system
		stateManager.makeUnbound(ce);
		//Use the current state as reference state
		stateManager.setReferenceState(ce, state);
	}

	public void refresh(WorkingSet ws, Object oid) throws PersistenceException {
		assertOpenWorkingset(ws);
		CacheEntry ce = cache.lookup(oid);
		if (ce != null) {
			refresh(ws, ce);
		}
	}
	public void refresh(WorkingSet ws, CacheEntry ce) throws PersistenceException {
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "refresh " + ce.getCeIdentifier());
		}
	    assertOpenWorkingset(ws);
		// Remove the reference state in order to reload from the database 
		// on the next access 
		stateManager.setReferenceState(ce, null);
		// Trash the current state associated to the working set 
		// (maybe not already attached)
		ws.bind(VirtualState.instance, 
		        ce.getCeIdentifier(), 
		        WorkingSet.UNKNOWN_INTENTION);
	}

	/**
	 * creates a working set
	 * @param userObject can be an external user depending on the
	 * personnality.
	 * @return a new WorkingSet
	 * @throws PersistenceException
	 */
	public WorkingSet createWS(Object userObject)
			throws PersistenceException {
		WorkingSet ws = wsManager.createWS(userObject);
		storage.beginWS(ws);
		return ws;
	}

	/**
	 * creates a working set
	 * @param userObject can be an external user depending on the
	 * personnality.
	 * @param workingSetType is a type of the working set.
	 * @return a new WorkingSet
	 * @throws PersistenceException
	 */
	public WorkingSet createWS(Object userObject,
							   Object workingSetType)
			throws PersistenceException {
		WorkingSet ws = wsManager.createWS(userObject, workingSetType);
		storage.beginWS(ws);
		return ws;
	}

	/**
	 * Closes and releases a context.
	 * @param context which must be closed
	 * @throws PersistenceException if the context is a transactional context
	 * which has not been commited or aborted.
	 */
	public void close(WorkingSet context) throws PersistenceException {
		if (!context.entries().isEmpty()) {
			boolean prepared_ok = false;
			PersistenceException pe = null;
			try {
				prepared_ok = prepareWS(context, true);
			} catch (PersistenceException e) {
				pe = e;
			}
			if (prepared_ok) {
				commitWS(context, WorkingSetLifeCycle.CLOSE_ACTION);
			} else {
				rollbackWS(context, WorkingSetLifeCycle.CLOSE_ACTION);
				if (pe == null) {
					throw new RolledBackPersistenceException();
				} else {
					throw pe;
				}
			}
		} else {
			context.getConnectionHolder().closeCHConnection();
			context.setStatus(WorkingSetLifeCycle.getNextStatus(
					context.getStatus(), WorkingSetLifeCycle.CLOSE_ACTION));
			storage.endWS(context);
		}
		wsManager.closeWS(context);
	}

	// IMPLEMENTATION OF THE TransactionalPersistenceManager INTERFACE //
	//-----------------------------------------------------------------//

	/**
	 * converts a context to a transaction
	 * @param tx becomes transactional
	 * @throws PersistenceException if the context is already transactional.
	 */
	public void begin(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG, "begin " + tx);
		}
		// 1st step: flush objects in working set, outside transaction
		if (tx.entries().size() > 0) {
			logger.log(BasicLevel.DEBUG, "Flush modification before the transaction " + tx);
			close(tx);
		}
		storage.beginWS(tx);
		tx.getConnectionHolder().begin();

		// 2nd step: change the transaction's status
		tx.setStatus(WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.BEGIN_TX_ACTION));
	}

	/**
	 * Prepare the transactinal context.
	 * @param tx the
	 * @return true if the transaction can be commit or false is the transaction
	 * must be rolled back.
	 * @throws PersistenceException if an error raised during the prepare
	 * operation, or if the context is not a transactional context.
	 */
	public boolean prepare(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
		return prepareWS(tx, !tx.getWSRollBackOnly());
	}

	/**
	 * commits a transactional context. The call to the prepare method is
	 * required.
	 * @param tx
	 * @throws PersistenceException if an error raised during the commit
	 * operation, or if the context is not a transactional context.
	 */
	public void commit(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
        try {
    		if (tx.getStatus() == TransactionalWorkingSet.CTX_ACTIVE_TRANSACTIONAL) {
    			//One phase commit
    			boolean prepared_ok = false;
    			PersistenceException pe = null;
    			try {
    				prepared_ok = prepareWS(tx, !tx.getWSRollBackOnly());
    			} catch (PersistenceException e) {
    				pe = e;
    			}
    			if (prepared_ok) {
    				commitWS(tx, WorkingSetLifeCycle.COMMIT_ACTION);
    			} else {
    				rollbackWS(tx, WorkingSetLifeCycle.ROLLBACK_ACTION);
    				if (pe == null) {
    					throw new RolledBackPersistenceException();
    				} else {
    					throw pe;
    				}
    			}
    		} else {
    			commitWS(tx, WorkingSetLifeCycle.COMMIT_ACTION);
            }
		} finally {
		    storage.beginWS(tx);
        }
	}

	/**
	 * rollbacks a transactional context. The call to the prepare method is
	 * required.
	 * @param tx
	 * @throws PersistenceException if an error raised during the rollback
	 * operation, or if the context is not a transactional context.
	 */
	public void rollback(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
        try {
    		//	check first the status of the working set
    		byte status = tx.getStatus(); 
    		if (status != TransactionalWorkingSet.CTX_PREPARED_FAIL && status != TransactionalWorkingSet.CTX_PREPARED_OK) {
    			try {
    				prepareWS(tx, false);
    			} catch (PersistenceException e) {
    				logger.log(BasicLevel.WARN, "An error occris during the rollback operation", e);
    			}
    		}
    		rollbackWS(tx, WorkingSetLifeCycle.ROLLBACK_ACTION);
        } finally {
            storage.beginWS(tx);
        }
	}

	//IMPLEMENTATION OF THE TPMAttributeControler INTERFACE //
	//--------------------------------------------------------//
	public void setFilters(Collection fs) throws IllegalArgumentException {
		this.filters = fs; 
	}
	
	/**
	 * @return the list of filters applied when ordering working set
	 */
	public Collection getFilters() {
		return filters;
	}
}
